//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "FormMain.h"
#include "FormAbout.h"
#include "FormSettings.h"
#include "Settings.h"
#include "LogUnit.h"
#include "Log.h"
#include "utils.h"
#include "Process.h"

#include <FileCtrl.hpp>

#include <memory>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMain *frmMain;
//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner)
	: TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".json" );
	appSettings.Read(asConfigFile);
	if (this->BorderStyle != bsSingle)
	{
		this->Width = appSettings.frmMain.iWidth;
		this->Height = appSettings.frmMain.iHeight;
	}
	this->Top = appSettings.frmMain.iPosY;
	this->Left = appSettings.frmMain.iPosX;
	if (appSettings.frmMain.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.frmMain.bWindowMaximized)
		this->WindowState = wsMaximized;
	if (appSettings.Logging.bLogToFile)
		CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
	else
		CLog::Instance()->SetFile("");
		
	edProjectDir->Text = appSettings.cppCommentRemover.dir;
	memoText->Text = appSettings.cppCommentRemover.textToRemove;
	for (int i=1; i<chlboxFiles->Count; i++)
	{
		chlboxFiles->Checked[i] = true;
	}
	cbMode->ItemIndex = appSettings.cppCommentRemover.mode;
	cbModeChange(NULL);	

	chbSkipLongComments->Checked = appSettings.cppCommentRemover.skipLongComments;
	edMaxCommentLength->Text = appSettings.cppCommentRemover.maxCommentLength;
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".json" );
	appSettings.frmMain.bWindowMaximized = (this->WindowState == wsMaximized);
	if (!appSettings.frmMain.bWindowMaximized)
	{
		// these values are meaningless is wnd is maximized
		appSettings.frmMain.iWidth = this->Width;
		appSettings.frmMain.iHeight = this->Height;
		appSettings.frmMain.iPosY = this->Top;
		appSettings.frmMain.iPosX = this->Left;
	}
	appSettings.cppCommentRemover.dir = edProjectDir->Text;
	appSettings.cppCommentRemover.textToRemove = memoText->Text;
	appSettings.cppCommentRemover.mode = static_cast<Settings::_CppCommentRemover::Mode>(cbMode->ItemIndex);

	appSettings.cppCommentRemover.skipLongComments = chbSkipLongComments->Checked;
	appSettings.cppCommentRemover.maxCommentLength = StrToIntDef(edMaxCommentLength->Text, appSettings.cppCommentRemover.maxCommentLength);

	appSettings.Write(asConfigFile);

	CanClose = true;	
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::actShowAboutExecute(TObject *Sender)
{
	frmAbout->ShowModal();	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::actShowSettingsExecute(TObject *Sender)
{
	frmSettings->appSettings = &appSettings;
	frmSettings->ShowModal();
	if (appSettings.frmMain.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.Logging.bLogToFile)
		CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
	else
		CLog::Instance()->SetFile("");
	frmLog->SetLogLinesLimit(appSettings.Logging.iMaxUiLogLines);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormShow(TObject *Sender)
{
    static bool once = false;
    if (!once)
    {
		once = true;
		frmLog->SetLogLinesLimit(appSettings.Logging.iMaxUiLogLines);		
		CLog::Instance()->SetLevel(E_LOG_TRACE);
		CLog::Instance()->callbackLog = frmLog->OnLog;
	}
	LOG("Application started\n");
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormDestroy(TObject *Sender)
{
	CLog::Instance()->Destroy();	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::actShowLogExecute(TObject *Sender)
{
	if (!frmLog->Visible)
		frmLog->Show();
	frmLog->BringToFront();
}
//---------------------------------------------------------------------------

bool MatchPattern(char *element, char *pattern)
{
	if( 0 == StrComp(pattern, "*") )
		return true;

	if( (*element == '\0') && (*pattern != '\0') )
		return false;

	if( *element == '\0' )
		return true;

	switch( *pattern )
	{
        case '*':
            if( MatchPattern(element, &pattern[1]) )
                return true;
            else
                return MatchPattern(&element[1], pattern);

        case '?':
            return MatchPattern(&element[1], &pattern[1]); 
    }

    if( *element == *pattern )
        return MatchPattern(&element[1], &pattern[1]);
    else
        return false;
}

void TfrmMain::ProcessDir(AnsiString path)
{
    LOG("Process directory: %s\n", path.c_str());
	int iAttributes = faAnyFile;
    Sysutils::TSearchRec sr;
	if (FindFirst(path, iAttributes, sr) == 0)
    {
        do
        {
            // drill into subdirs
            if (sr.Attr & faDirectory) {
                // Do not drill into "." or ".."
                if (sr.Name != "." && sr.Name != "..") {
					if(path.Length() + sr.Name.Length() < MAX_PATH-1)
					{
						AnsiString drillPath = appendDirectory(path, sr.Name);
                        ProcessDir(drillPath);
                    }
                }
            }
            else
            {
				for (int i=0; i<chlboxFiles->Items->Count; i++)
				{
					if (chlboxFiles->Checked[i])
					{
						if (MatchPattern(sr.Name.c_str(), chlboxFiles->Items->Strings[i].c_str()))
						{
							AnsiString Path = ExtractFilePath(path) + sr.Name;
							ProcessFile(Path);
						}
					}
				}
			}
        } while (FindNext(sr) == 0);
        FindClose(sr);
	}
}

void TfrmMain::ProcessFile(AnsiString name)
{
	LOG("    Process: %s\n", name.c_str());

	std::auto_ptr<TStrings> strings(new TStringList());

	try
	{
		strings->LoadFromFile(name);
	}
	catch(...)
	{
		LOG("Failed to load data from %s file\n", name.c_str());
		return;
	}
	AnsiString text = strings->Text;
	AnsiString out;

	if (ProcessCpp	(	text,
						memoText->Lines,
						(cbMode->ItemIndex == Settings::_CppCommentRemover::DELETE_ALL),
						chbSkipLongComments->Checked,
						StrToIntDef(edMaxCommentLength->Text, appSettings.cppCommentRemover.maxCommentLength),
						out
					)
		)
	{
		return;
	}

	if (text == out)
	{
		// not modified
		return;
    }

	FILETIME ftCreate, ftAccess, ftWrite;
	bool timestampsValid = false;
	HANDLE hFile = INVALID_HANDLE_VALUE;
	if (appSettings.cppCommentRemover.preserveTimestamps)
	{
		hFile = CreateFile(name.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		// Retrieve the file times for the file.
		if (hFile != INVALID_HANDLE_VALUE)
		{
			if (!GetFileTime(hFile, &ftCreate, &ftAccess, &ftWrite))
			{
				LOG("Failed to retrieve time stamps for %s\n", name.c_str());
			}
			else
			{
				timestampsValid = true;
			}
			CloseHandle(hFile);
			hFile = INVALID_HANDLE_VALUE;
		}
		else
		{
			LOG("Failed to open file for reading (fetching timestamps)\n");
		}
	}

	std::auto_ptr<TStrings> stringsOut(new TStringList());
	stringsOut->Text = out;
	try
	{
		stringsOut->SaveToFile(name);
		if (timestampsValid)
		{
			hFile = CreateFile(name.c_str(), GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
			if (hFile != INVALID_HANDLE_VALUE)
			{
				if (!SetFileTime(hFile, &ftCreate, &ftAccess, &ftWrite))
				{
					LOG("Failed to set timestamps for %s\n", name.c_str());
				}
				CloseHandle(hFile);
				hFile = INVALID_HANDLE_VALUE;
			}
			else
			{
				LOG("Failed to open file for writing timestamps\n");
            }			
        }
	}
	catch(...)
	{
		LOG("Failed to save %s\n", name.c_str());
	}
}

void __fastcall TfrmMain::btnBrowseDirClick(TObject *Sender)
{
    AnsiString asDir;
	if (DirectoryExists(edProjectDir->Text))
		asDir = edProjectDir->Text;
	if (SelectDirectory("Project dir:", "", asDir))
		edProjectDir->Text = asDir;	
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::btnDeleteCommentsClick(TObject *Sender)
{
	if (appSettings.cppCommentRemover.disableWarning == false)
	{
		if (MessageBox(this->Handle,
			"Files would be processed in place (overwritten).\r\n"
			"Things can happen - make sure you have backup or working on copy.\r\n"
			"Are you sure you want to continue?",
			this->Caption.c_str(), MB_YESNO | MB_DEFBUTTON2 | MB_ICONEXCLAMATION) != IDYES)
		{
			return;
		}
	}
	btnDeleteComments->Enabled = false;
	AnsiString tmp = btnDeleteComments->Caption;
	btnDeleteComments->Caption = "Processing...";
	ProcessDir(edProjectDir->Text + "\\*.*");
	btnDeleteComments->Enabled = true;
	btnDeleteComments->Caption = tmp;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnTestClick(TObject *Sender)
{
	AnsiString out;
	if (ProcessCpp(memoIn->Text, memoText->Lines, (cbMode->ItemIndex == Settings::_CppCommentRemover::DELETE_ALL),
		chbSkipLongComments->Checked,
		StrToIntDef(edMaxCommentLength->Text, appSettings.cppCommentRemover.maxCommentLength),
		out))
	{
		memoOut->Text = "";
		return;
	}
	memoOut->Text = out;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::cbModeChange(TObject *Sender)
{
	memoText->Visible = (cbMode->ItemIndex == 0);	
}
//---------------------------------------------------------------------------

